// Copyright (c) 2006  John Abbott
// This file is part of the CoCoALib suite of examples.
// You are free to use any part of this example in your own programs.

#include "CoCoA/library.H"
using namespace CoCoA;
using namespace std;

//----------------------------------------------------------------------
const string ShortDescription =
  "Program exhibiting a way of using ever higher precisions...\n"
  "This example shows how failure can be a success: this computation\n"
  "is proven to give a wrong result for any given precision!";

const string LongDescription =
  "Example showing iterative increase of precision using RingTwinFloat    \n"
  "until the answer is found (or an arbitrary limit on precision is reached).\n"
  "This program will always fail to find the limit: J-M Muller's sequence \n"
  "actually converges to 6 (rather slowly), however it is pathological    \n"
  "because it converges to 100 using any finite precision arithmetic.     \n"
  "RingTwinFloat detects the onset of pathological convergence to 100, and\n"
  "throws an InsufficientPrecision exception.                             \n";
//----------------------------------------------------------------------

// Includes from the standard C++ library
// #include <iostream> // using std::endl;
#include <algorithm> // using std::swap;


// This function computes J-M Muller's sequence in the ring R until two
// successive values are "practically equal" (see documentation).
// It returns the value the sequence converged to.
// Note: the initial values are 4 and 7/2 which can be represented exactly.
RingElem MullerSeq(const RingTwinFloat& R)
{
  RingElem Vprev1(R, 7); Vprev1 /= 2; // These starting values can be represented
  RingElem Vprev2(R, 4);              // exactly (on a binary computer).

  RingElem Vcurr(R);
  while (!IsPracticallyEqual(Vprev1, Vprev2))
  {
    Vcurr = 111 - 1130/Vprev1 + 3000/(Vprev1*Vprev2); // Muller's recurrence.
    Vprev2 = Vprev1;
    Vprev1 = Vcurr;
  }
  return Vprev1;
}


// This loop tries ever higher precisions until the answer is found.
// Arbitrarily limit precision to 16384 so it doesn't run for ever.
RingElem MullerLoop()
{
  for (size_t BitPrec = 32; BitPrec <= 8192; BitPrec *= 2)
  {
    try
    {
      return MullerSeq(NewRingTwinFloat(BitPrec));
    }
    catch (RingTwinFloat::InsufficientPrecision&)
    {
      // Inform user about the failure...
      GlobalLogput() << "A bit precision of " << BitPrec << " was not sufficient." << endl;
    }
  }
  CoCoA_ERROR("Required too much precision... giving up!", "MullerLoop");
  return zero(RingZ()); // Never executed; just to keep compiler quiet.
}


void program()
{
  GlobalManager CoCoAFoundations;

  GlobalOutput() << ShortDescription << endl;

  try
  {
    RingElem limit = MullerLoop();
    GlobalOutput() << "Muller's sequence converges to " << limit << endl;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    if (string(err.what()) != "Required too much precision... giving up!") throw;
    GlobalOutput() << "As expected we caught the error: " << err.what() << endl;
  }
}


//----------------------------------------------------------------------
// Use main() to handle any uncaught exceptions and warn the user about them.
int main()
{
  try
  {
    program();
    return 0;
  }
  catch (const CoCoA::ErrorInfo& err)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT CoCoA error";
    ANNOUNCE(err);
  }
  catch (const std::exception& exc)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT std::exception: " << exc.what() << endl;
  }
  catch(...)
  {
    GlobalErrput() << "***ERROR***  UNCAUGHT UNKNOWN EXCEPTION" << endl;
  }
  return 1;
}

//----------------------------------------------------------------------
// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/examples/ex-RingTwinFloat2.C,v 1.1 2007/06/21 21:29:47 abbott Exp $
// $Log: ex-RingTwinFloat2.C,v $
// Revision 1.1  2007/06/21 21:29:47  abbott
// Changed name of RingFloat into RingTwinFloat.
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.7  2007/03/06 13:04:06  bigatti
// -- cleaned: it catches the error it throws and performs 1 fewer operation
//
// Revision 1.6  2007/03/03 14:15:45  bigatti
// -- "foundations" renamed into "GlobalManager"
//
// Revision 1.5  2007/02/22 17:24:32  bigatti
// -- added a comment in ShortDescription
//
// Revision 1.4  2007/02/12 15:59:00  bigatti
// -- added strings ShortDescription and LongDescription for indexing
//
// Revision 1.3  2007/02/10 18:44:04  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.2  2006/12/21 13:48:33  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.2  2006/02/20 22:41:20  cocoa
// All forms of the log function for power products now return SmallExponent_t
// (instead of int).  exponents now resizes the vector rather than requiring
// the user to pass in the correct size.
//

