#ifndef CoCoA_matrix_H
#define CoCoA_matrix_H

//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "CoCoA/ring.H"
#include "CoCoA/ZZ.H"

#include <iosfwd>
//using std::ostream;
#include <cstddef>
//using std::size_t;
#include <vector>
//using std::vector;


namespace CoCoA
{

  class OpenMathOutput; // forward declaration -- defined in OpenMath.H

//   class MatrixVectorBase
//   {
//   public:
//     ~MatrixVectorBase(); // needed by inherited classes
//     std::size_t myNumCompts() = 0;
//     RingRawValue& myRawEntry(std::size_t i) = 0;
//     void myEntry(RingElem& r, std::size_t i) = 0;
//     void mySetEntry(std::size_t i, const RefRingElem& r) = 0;
//   };

//   class MatrixVector
//   {
//   public:
//     MatrixVectorBase* operator->();
//     RingElem operator[](std::size_t i);
//     friend const ring& owner(const MatrixVector& v);
//     friend void set(MatrixVector& v, std::size_t i, const RefRingElem& r);
//   private:
//     ring myR;
//     MatrixVectorBase& myCoords;
//   };


  /*-----------------------------------------------------------------*/
  /** \include matrix.txt  */
  /*-----------------------------------------------------------------*/


  class MatrixBase; // forward declaration
  class ConstMatrixBase; // forward declaration

  class ConstMatrix
  {
    // smart ref counted pointer class???
  public:
    explicit ConstMatrix(const ConstMatrixBase* MatPtr); // public???
    ConstMatrix(const ConstMatrix& M); // copy ctor
    ~ConstMatrix();  // ??? should this be virtual ????
    const ConstMatrixBase* myMatrixPtr() const;
    const ConstMatrixBase* operator->() const;
    ConstRefRingElem operator()(std::size_t i, std::size_t j) const;
  private: // disable assignment
    ConstMatrix& operator=(const ConstMatrix&); //  NEVER DEFINED -- ASSIGMENT DISABLED
  protected: // data member
    ConstMatrixBase const* myPtr;
  };


  class matrix: public ConstMatrix
  {
    // Data members inherited from ConstMatrix;
    // will actually cast the pointer to MatrixBase*
  public:
    matrix(MatrixBase* MatPtr); // public???
    matrix(const matrix& copy); // copy ctor (not a deep copy)
    // default dtor works fine
    MatrixBase* myMatrixPtr();
    MatrixBase* operator->();
  private: // disable assignment and defauilt copy ctor
    matrix& operator=(const matrix&); //  NEVER DEFINED -- assigment disabled
  };



  std::ostream& operator<<(std::ostream& out, ConstMatrix M);
  void AssignZero(matrix& M);
  void SetEntry(matrix& M, std::size_t i, std::size_t j, ConstRefRingElem r);
  void SetEntry(matrix& M, std::size_t i, std::size_t j, long n);
  void SetEntry(matrix& M, std::size_t i, std::size_t j, const ZZ& N);
  bool IsZeroRow(ConstMatrix M, std::size_t i);
  bool IsZeroCol(ConstMatrix M, std::size_t j);
  RingElem det(ConstMatrix M);
  std::size_t rank(ConstMatrix M);




  class ConstMatrixBase
  {
  private: // data members
    mutable std::size_t myRefCount; // every matrix has an intrusive reference count
  protected: // ref count fns accessible only to friends and derived classes.
    friend class ConstMatrix;
//???    friend class matrix;      // so that matrix::~matrix() can call the dtor
    void myRefCountInc() const;
    void myRefCountDec() const;
    void myRefCountZero() const;

  protected:
    ConstMatrixBase();  // inline fn, just zeroes myRefCount
    virtual ~ConstMatrixBase(); // defined out of line, does nothing
  public:
    typedef std::vector<RingElem> vec;
    virtual const ring& myBaseRing() const = 0;
    virtual std::size_t myNumRows() const = 0;
    virtual std::size_t myNumCols() const = 0;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const = 0;
    virtual void myMulByRow(vec& lhs, const vec& v) const; ///< lhs = v*M -- vector-by-matrix product, default "dense" definition
    virtual void myMulByCol(vec& lhs, const vec& v) const; ///< lhs = M*v -- matrix-by-vector product, default "dense" definition
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero -- default "dense" definition
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero -- default "dense" definition
    virtual void myDet(RefRingElem d) const;                      // default method uses Gaussian reduction
    virtual size_t myRank() const;                                // default method uses Gaussian reduction
    virtual void myOutputSelf(std::ostream& out) const;           // has a default "dense" definition
    virtual void myOutputSelf(OpenMathOutput& OMOut) const; // has a default "dense" definition
    virtual void myCheckIndices(std::size_t i, std::size_t j, const char* where="") const;  // throws if indices are too large
  };



  class MatrixBase: public ConstMatrixBase
  {
    // Data members: myRefCount inherited from ConstMatrixBase

  protected:
    MatrixBase();
    friend class matrix; // so that matrix::~matrix() can call the dtor
    virtual ~MatrixBase();
  public:
    virtual MatrixBase* myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const = 0;
    typedef std::vector<RingElem> vec;
    virtual void mySwap(matrix& M) = 0; // homogeneous swap
    virtual void myAssignZero() = 0; // sets all entries to zero
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r) = 0;
    virtual void mySetEntry(std::size_t i, std::size_t j, long n) = 0;
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N) = 0;
    virtual void myRowMul(std::size_t i, ConstRefRingElem c) = 0; ///< row(i) = c*row(i)
    virtual void myColMul(std::size_t j, ConstRefRingElem c) = 0; ///< row(j) = c*row(j)
    virtual void myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c) = 0; ///< row(i1) += c*row(i2)
    virtual void myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c) = 0; ///< row(j1) += c*row(j2)
    virtual void mySwapRows(std::size_t i1, std::size_t i2) = 0;
    virtual void mySwapCols(std::size_t j1, std::size_t j2) = 0;
  };





  /////////////////////////////////////////////////////////////////////////////
  // Inline functions

  inline void ConstMatrixBase::myRefCountInc() const { ++myRefCount; }
  inline void ConstMatrixBase::myRefCountDec() const { if (--myRefCount == 0) delete this; }
  inline void ConstMatrixBase::myRefCountZero() const { myRefCount = 0; }


//   inline MatrixVectorBase* MatrixVector::operator->()
//   {
//     return &myCoords;
//   }


//   inline const ring& owner(const MatrixVector& v)
//   {
//     return v.myR;
//   }


  inline ConstMatrix::ConstMatrix(const ConstMatrixBase* MatPtr):
      myPtr(MatPtr)
  {
    myPtr->myRefCountInc();
  }


  inline ConstMatrix::ConstMatrix(const ConstMatrix& copy):
      myPtr(copy.myPtr)
  {
    myPtr->myRefCountInc();
  }

  inline ConstMatrix::~ConstMatrix()
  {
    myPtr->myRefCountDec();
  }


  inline const ConstMatrixBase* ConstMatrix::myMatrixPtr() const
  {
    return myPtr;
  }


  inline const ConstMatrixBase* ConstMatrix::operator->() const
  {
    return myPtr;
  }


  // Currently ConstMatrix::operator() is out of line -- see matrix.C


  inline matrix::matrix(MatrixBase* MatPtr):
      ConstMatrix(MatPtr)
  {}


  inline matrix::matrix(const matrix& copy):
      ConstMatrix(copy)
  {}


  inline MatrixBase* matrix::myMatrixPtr()
  {
    // dynamic_cast would perform needless checks
    // reinterpret_cast might not be portable?
    return static_cast<MatrixBase*>(const_cast<ConstMatrixBase*>(myPtr));
  }


  inline MatrixBase* matrix::operator->()
  {
    // dynamic_cast would perform needless checks
    // reinterpret_cast might not be portable?
    return static_cast<MatrixBase*>(const_cast<ConstMatrixBase*>(myPtr));
  }


  inline const ring& BaseRing(ConstMatrix M)
  {
    return M->myBaseRing();
  }


  inline std::size_t NumRows(ConstMatrix M)
  {
    return M->myNumRows();
  }


  inline std::size_t NumCols(ConstMatrix M)
  {
    return M->myNumCols();
  }


} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/matrix.H,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: matrix.H,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2006/12/06 17:18:40  cocoa
// -- removed #include "config.H"
//
// Revision 1.4  2006/11/24 17:41:18  cocoa
// -- reorganized includes of header files
//
// Revision 1.3  2006/11/02 13:25:44  cocoa
// Simplification of header files: the OpenMath classes have been renamed.
// Many minor consequential changes.
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:36  cocoa
// Imported files
//
// Revision 1.4  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.3  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.2  2006/03/27 12:21:26  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:30  cocoa
// Imported files
//
// Revision 1.11  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.10  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.9  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.8  2005/03/31 16:59:42  cocoa
// Made special matrix ctors private, so a user has to pass via the
// pseudo-ctors (which do all the arg sanity checking).
//
// Revision 1.7  2005/03/30 17:15:14  cocoa
// Cleaned the SpecialMatrix code; a simple test compiles and
// runs fine.
//
// Revision 1.6  2005/03/29 17:36:47  cocoa
// Just checking in before going home -- test-matrix1 does not yet link!
//
// Revision 1.5  2005/03/29 17:19:30  cocoa
// -- added: rank and myRank
//
// Revision 1.4  2005/03/11 16:44:18  cocoa
// New abstract class structure for matrices.
// New types of special matrix.
//
// Revision 1.3  2005/03/02 18:46:41  cocoa
// Added new types ConstRefMatrix, and RefMatrix following along
// the lines of ConstRefRingElem and RefRingElem.  The semantics
// should be a bit clearer now.
//
// Revision 1.2  2005/02/11 14:15:20  cocoa
// New style ring elements and references to ring elements;
// I hope I have finally got it right!
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.9  2004/11/29 16:19:02  cocoa
// -- changed syntax for function det
//
// Revision 1.8  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//

#endif
