#ifndef CoCoA_symbol_H
#define CoCoA_symbol_H

//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "CoCoA/assert.H"

#include <cstddef>
//using std::size_t;
#include <vector>
//using std::vector;
#include <string>
//using std::string;
#include <iosfwd>
//using std::ostream;


namespace CoCoA
{

  class symbol
  {
  public:
    explicit symbol(const std::string& head); // symbolic name without indices
    symbol(const std::string& head, int index); // symbolic name with 1 index
    symbol(const std::string& head, const std::vector<int>& indices); // symbolic name with indices
    // Default copy ctor and assignment are fine.
    // Default dtor is fine.

    void myOutputSelf(std::ostream& out) const;
    void myInput(std::istream& in);

    int myCmp(const symbol& sym2) const;
    static bool IsValidHead(const std::string& head);
  private: // data members
    std::string myHead;
    std::vector<int> myIndices;
    friend const std::string& head(const symbol& sym);
    friend std::size_t NumIndices(const symbol& sym);
    friend int index(const symbol& sym, size_t n);
  };


  const std::string& head(const symbol& sym);                   // inline
  std::size_t NumIndices(const symbol& sym);                    // inline
  int index(const symbol& sym, size_t n);
  bool AreDistinct(const std::vector<symbol>& syms);
  bool AreArityConsistent(const std::vector<symbol>& syms);
  std::ostream& operator<<(std::ostream& out, const symbol& sym);
  std::istream& operator>>(std::istream& in, symbol& sym);
  int cmp(const symbol& sym1, const symbol& sym2);              // inline
  bool operator==(const symbol& sym1, const symbol& sym2);      // inline
  bool operator!=(const symbol& sym1, const symbol& sym2);      // inline
  bool operator<(const symbol& sym1, const symbol& sym2);       // inline
  bool operator<=(const symbol& sym1, const symbol& sym2);      // inline
  bool operator>(const symbol& sym1, const symbol& sym2);       // inline
  bool operator>=(const symbol& sym1, const symbol& sym2);      // inline

  std::vector<symbol> SymbolRange(const symbol& sym1, const symbol& sym2);
  std::vector<symbol> SymbolRange(const std::string& head, int lo, int hi); // <==> SymbolRange(symbol(head,lo),symbol(head,hi));
  // Four convenience functions -- make creating simple PolyRings easier.
  std::vector<symbol> symbols(const std::string& head1);
  std::vector<symbol> symbols(const std::string& head1, const std::string& head2);
  std::vector<symbol> symbols(const std::string& head1, const std::string& head2, const std::string& head3);
  std::vector<symbol> symbols(const std::string& head1, const std::string& head2, const std::string& head3, const std::string& head4);

  //---------------------------------------------------------------------------
  // Inline implementations

  inline const std::string& head(const symbol& sym)
  {
    return sym.myHead;
  }


  inline std::size_t NumIndices(const symbol& sym)
  {
    return sym.myIndices.size();
  }


  inline int cmp(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2);
  }

  inline bool operator==(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) == 0;
  }

  inline bool operator!=(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) != 0;
  }

  inline bool operator<(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) < 0;
  }

  inline bool operator<=(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) <= 0;
  }

  inline bool operator>(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) > 0;
  }

  inline bool operator>=(const symbol& sym1, const symbol& sym2)
  {
    return sym1.myCmp(sym2) >= 0;
  }

} // end of namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/include/CoCoA/symbol.H,v 1.3 2007/05/31 14:50:09 bigatti Exp $
// $Log: symbol.H,v $
// Revision 1.3  2007/05/31 14:50:09  bigatti
// -- added AreDistinct and AreArityConsistent
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.6  2007/03/08 16:55:06  cocoa
// Changed name of "range" function to "SymbolRange".
//
// Revision 1.5  2007/03/08 14:38:07  cocoa
// Added new range function in symbol.H, and tidied many calls to PolyRing
// pseudo ctors (as a consequence).
//
// Revision 1.4  2006/11/27 14:20:09  cocoa
// -- removed #include "CoCoA/config.H"
//
// Revision 1.3  2006/10/27 19:09:45  cocoa
// Replaced some member functions of CoCoA::symbol by friend functions.
// Removed some include dependency on symbol.H
//
// Revision 1.2  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.2  2006/03/15 18:09:31  cocoa
// Changed names of member functions which print out their object
// into myOutputSelf -- hope this will appease the Intel C++ compiler.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1  2005/07/08 15:09:29  cocoa
// Added new symbol class (to represent names of indets).
// Integrated the new class into concrete polynomial rings
// and PPMonoid -- many consequential changes.
// Change ctors for the "inline" sparse poly rings: they no
// longer expect a PPMonoid, but build their own instead
// (has to be a PPMonoidOv).
//

#endif
