//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "CoCoA/DiagMatrix.H"
#include "CoCoA/assert.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"

//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;
#include <algorithm>
using std::min;
using std::fill;
//using std::swap;

namespace CoCoA
{

  // DiagMatrixImpl -- diagonal matrices implemented in a pretty obvious manner

  class DiagMatrixImpl: public MatrixBase
  {
  public:
    DiagMatrixImpl(const ring& R, std::size_t nrows, std::size_t ncols); // creates zero matrix
    DiagMatrixImpl(const ring& R, std::size_t nrows, std::size_t ncols, const std::vector<RingElem>& DiagEntries);
    ~DiagMatrixImpl();

  public:
//???    MatrixBase* clone() const = 0;
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual DiagMatrixImpl* myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const;
    virtual void mySwap(matrix& M); // homogeneous swap
    virtual void myAssignZero(); // sets all entries to zero
    virtual void mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r);
    virtual void mySetEntry(std::size_t i, std::size_t j, long n);
    virtual void mySetEntry(std::size_t i, std::size_t j, const ZZ& N);
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myRowMul(std::size_t i, ConstRefRingElem c); ///< row(i) = c*row(i)
    virtual void myColMul(std::size_t j, ConstRefRingElem c); ///< row(j) = c*row(j)
    virtual void myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c); ///< row(i1) += c*row(i2)
    virtual void myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c); ///< row(j1) += c*row(j2)
    virtual void mySwapRows(std::size_t i1, std::size_t i2);
    virtual void mySwapCols(std::size_t j1, std::size_t j2);
    virtual void myDet(RefRingElem d) const;
    virtual size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    std::vector<RingElem> myDiagEntries; //??? why not vector<RingElemRawPtr>???
  };

  //----------------------------------------------------------------------

  DiagMatrixImpl::DiagMatrixImpl(const ring& R, std::size_t NumRows, std::size_t NumCols):
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols)
  {
    CoCoA_ASSERT(NumRows == NumCols); // ??? restrict to square diagonal matrices???
    const size_t n = min(NumRows, NumCols);
    myDiagEntries.resize(n, zero(R));
  }


  DiagMatrixImpl::DiagMatrixImpl(const ring& R, std::size_t NumRows, std::size_t NumCols, const std::vector<RingElem>& DiagEntries):
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols),
      myDiagEntries(DiagEntries)
  {
    CoCoA_ASSERT(NumRows == NumCols); // ??? restrict to square diagonal matrices???
    CoCoA_ASSERT(myDiagEntries.size() == min(NumRows, NumCols));
  }


  DiagMatrixImpl::~DiagMatrixImpl()
  {}


  const ring& DiagMatrixImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t DiagMatrixImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t DiagMatrixImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem DiagMatrixImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i == j)
      return myDiagEntries[i];
    else
      return zero(myR);
  }


  DiagMatrixImpl* DiagMatrixImpl::myZeroClone(const ring& R, std::size_t NumRows, std::size_t NumCols) const
  {
    return new DiagMatrixImpl(R, NumRows, NumCols);
  }


  void DiagMatrixImpl::mySwap(matrix& M)
  {
    DiagMatrixImpl* other = dynamic_cast<DiagMatrixImpl*>(M.myMatrixPtr());
    CoCoA_ASSERT(other != 0);
    CoCoA_ASSERT(myBaseRing() == BaseRing(M));
    std::swap(myNumRowsValue, other->myNumRowsValue);
    std::swap(myNumColsValue, other->myNumColsValue);
    std::swap(myDiagEntries, other->myDiagEntries);
  }


  // Exception clean only if the call to fill cannot throw
  void DiagMatrixImpl::myAssignZero()
  {
    fill(myDiagEntries.begin(), myDiagEntries.end(), zero(myR));
  }


  void DiagMatrixImpl::mySetEntry(std::size_t i, std::size_t j, ConstRefRingElem r)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i != j)
    {
      CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySetEntry");
    }
    myDiagEntries[i] = r;
  }


  void DiagMatrixImpl::mySetEntry(std::size_t i, std::size_t j, long n)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i != j)
    {
      CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySetEntry");
    }
    myDiagEntries[i] = n;
  }


  void DiagMatrixImpl::mySetEntry(std::size_t i, std::size_t j, const ZZ& N)
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i != j)
    {
      CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySetEntry");
    }
    myDiagEntries[i] = N;
  }


  void DiagMatrixImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    vector<RingElem> ans(myNumCols(), zero(myR));
    const size_t n = min(myNumRows(), myNumCols());
    for (size_t i=0; i < n; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    std::swap(lhs, ans); //??? does this clobber iterators on lhs???
  }


  void DiagMatrixImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    vector<RingElem> ans(myNumRows(), zero(myR));
    const size_t n = min(myNumRows(), myNumCols());
    for (size_t i=0; i < n; ++i)
      ans[i] = v[i]*myDiagEntries[i];
    std::swap(lhs, ans); //??? does this clobber iterators on lhs???
  }


  bool DiagMatrixImpl::myIsZeroRow(std::size_t i) const
  {
    if (i >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DiagMatrixImpl::myIsZeroRow");
    if (i >= myNumCols()) return true;
    return IsZero(myDiagEntries[i]);
  }


  bool DiagMatrixImpl::myIsZeroCol(std::size_t j) const
  {
    if (j >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DiagMatrixImpl::myIsZeroCol");
    if (j >= myNumRows()) return true;
    return IsZero(myDiagEntries[j]);
  }


  void DiagMatrixImpl::myRowMul(std::size_t i, ConstRefRingElem c)
  {
    if (i >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DiagMatrixImpl::myRowMul");
    if (i >= myNumCols()) return;
    myDiagEntries[i] *= c;
  }


  void DiagMatrixImpl::myColMul(std::size_t j, ConstRefRingElem c)
  {
    if (j >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DiagMatrixImpl::myColMul");
    if (j >= myNumCols()) return;
    myDiagEntries[j] *= c;
  }


  void DiagMatrixImpl::myAddRowMul(std::size_t i1, std::size_t i2, ConstRefRingElem c)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::myAddRowMul"); // ???should always give error???
    if (i1 >= myNumRows() || i2 >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DiagMatrixImpl::myAddRowMul");
    if (IsZero(c) || i2 >= myNumCols() || IsZero(myDiagEntries[i2])) return;
    RingElem prod = c*myDiagEntries[i2];
    if (IsZero(prod)) return; // there might be zero divisors
    if (i1 != i2)
      CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::myAddRowMul");
    myDiagEntries[i1] += prod;
  }


  void DiagMatrixImpl::myAddColMul(std::size_t j1, std::size_t j2, ConstRefRingElem c)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::myAddColMul"); // ???should always give error???
    if (j1 >= myNumCols() || j2 >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DiagMatrixImpl::myAddColMul");
    if (IsZero(c) || j2 >= myNumRows() || IsZero(myDiagEntries[j2])) return;
    RingElem prod = c*myDiagEntries[j2];
    if (IsZero(prod)) return; // there might be zero divisors
    if (j1 != j2)
      CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::myAddColMul");
    myDiagEntries[j1] += prod;
  }


  void DiagMatrixImpl::mySwapRows(std::size_t i1, std::size_t i2)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySwapRows");  // ???should always give error???
    if (i1 >= myNumRows() || i2 >= myNumRows())
      CoCoA_ERROR(ERR::BadRowIndex, "DiagMatrixImpl::mySwapRows");
    if (i1 == i2) return;
    if (myIsZeroRow(i1) && myIsZeroRow(i2)) return;
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySwapRows");
  }


  void DiagMatrixImpl::mySwapCols(std::size_t j1, std::size_t j2)
  {
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySwapCols");  // ???should always give error???
    if (j1 >= myNumCols() || j2 >= myNumCols())
      CoCoA_ERROR(ERR::BadColIndex, "DiagMatrixImpl::mySwapCols");
    if (j1 == j2) return;
    if (myIsZeroCol(j1) && myIsZeroCol(j2)) return;
    CoCoA_ERROR(ERR::ConstMatEntry, "DiagMatrixImpl::mySwapCols");
  }


  void DiagMatrixImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(myNumRows() == myNumRows());
    RingElem ans(myR, 1);
    for (size_t i=0; i < myNumRows(); ++i)
      ans *= myDiagEntries[i];
    swap(d, ans);
  }


  // ??? better use "count"?
  size_t DiagMatrixImpl::myRank() const
  {
    size_t r=0;
    for (size_t i=min(myNumRows(),myNumCols()); i-- > 0;)
      if (!IsZero(myDiagEntries[i])) ++r;
    return r;
  }


  matrix NewDiagMatrix(const ring& R, std::size_t NumRows, std::size_t NumCols)
  {
    if (NumRows != NumCols)
      CoCoA_ERROR(ERR::NotSquareMatrix, "NewDiagMatrix(R, nrows, ncols)");
    return matrix(new DiagMatrixImpl(R, NumRows, NumCols));
  }


  matrix NewDiagMatrix(const ring& R, std::size_t NumRows, std::size_t NumCols, std::vector<RingElem> DiagEntries)
  {
    if (NumRows != NumCols)
      CoCoA_ERROR(ERR::NotSquareMatrix, "NewDiagMatrix(R, nrows, ncols, DiagEntries)");
    if (DiagEntries.size() != min(NumRows, NumCols))
      CoCoA_ERROR(ERR::BadArraySize, "NewDiagMatrix(R, nrows, ncols, DiagEntries)");
    for (size_t i=0; i < DiagEntries.size(); ++i)
      if (owner(DiagEntries[i]) != R)
        CoCoA_ERROR(ERR::MixedRings, "NewDiagMatrix(R, nrows, ncols, DiagEntries)");
    return matrix(new DiagMatrixImpl(R, NumRows, NumCols, DiagEntries));
  }

//???  matrix NewDiagMatrix(ConstMatrix M);


} // End of namespace CoCoA
