//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "CoCoA/MatrixArith.H"
#include "CoCoA/assert.H"
#include "CoCoA/matrix.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingZ.H"

#include <iostream>
using std::ostream;
//#include <cstddef>
using std::size_t;


namespace CoCoA
{


  matrix NewMatrixHElim(ConstMatrix GradM, std::vector<size_t> IndetsToElim);
  matrix NewMatrixAElim(ConstMatrix GradM, std::vector<size_t> IndetsToElim);
  matrix NewMatrix0Elim(ConstMatrix GradM, std::vector<size_t> IndetsToElim);


  void DetByGauss(RefRingElem d, ConstMatrix M)
  {
    ring R(BaseRing(M));
    // ??? this function works only within integral domains!
    CoCoA_ASSERT(IsIntegralDomain(R));
    CoCoA_ASSERT(NumRows(M) == NumRows(M));
    ring K((IsField(R) ? R : NewFractionField(R)));
    RingHom R2K(R==K ? IdentityHom(K) : EmbeddingHom(AsFractionField(K)));
    matrix Gss(NewDenseMatrix(K, NumRows(M), NumCols(M)));
    // below copy M via RingHom into Gss
    for (size_t row=0; row < NumRows(M); ++row)
      for (size_t col=0; col < NumCols(M); ++col)
        SetEntry(Gss, row, col, R2K(M(row,col)));
    RingElem c(K);
    RingElem determinant(one(K));
    for (size_t col=0; col < NumCols(M); ++col)
    {
      if (IsZero(Gss(col,col)))
      {
        size_t r=col+1;
        for ( ; r < NumRows(M); ++r)
          if (!IsZero(Gss(r,col)))
          {
            Gss->mySwapRows(r,col);
            determinant = -determinant;
            break;
          }
        if (r==NumRows(M))
        {
          d = zero(R);
          return;
        }
      }
      c = Gss(col,col);
      determinant *= c;
      for (size_t i=col+1; i < NumRows(M); ++i)
        Gss->myAddRowMul(i, col, -Gss(i,col)/c);
    }
    if (R==K) swap(d, determinant);
    else swap(d, num(determinant)/den(determinant));
  }


  size_t RankByGauss(ConstMatrix M)
  {
    // ??? this function works only within integral domains!
    ring R(BaseRing(M));
    if (!IsIntegralDomain(R)) CoCoA_ERROR(ERR::NYI, "RankByGauss over non-integral domain");
    ring K((IsField(R) ? R : NewFractionField(R)));
    RingHom R2K(R==K ? IdentityHom(K) : EmbeddingHom(AsFractionField(K)));
    matrix Gss(NewDenseMatrix(K, NumRows(M), NumCols(M)));
    // below copy M via RingHom into Gss
    {
      // This will eventually become a separate function
      for (size_t row=0; row < NumRows(M); ++row)
        for (size_t col=0; col < NumCols(M); ++col)
          SetEntry(Gss, row, col, R2K(M(row,col)));
    }
    RingElem c(K);
    size_t row=0;
    for (size_t col=0; col < NumCols(M); ++col)
    {
      if (IsZero(Gss(row,col)))
      {
        size_t i=row+1;
        for ( ; i < NumRows(M); ++i)
          if (!IsZero(Gss(i,col)))
          {
            Gss->mySwapRows(i,row);
            break;
          }
        if (i==NumRows(M)) continue;
      }
      c = Gss(row,col);
      for (size_t i=row+1; i < NumRows(M); ++i)
	Gss->myAddRowMul(i, row, -Gss(i,col)/c);
      ++row;
    }
    return row;
  }

  void InverseByGauss(matrix& lhs, ConstMatrix M)
  {
    // this code works only if the base ring is a field
    CoCoA_ASSERT(IsField(BaseRing(M)));
    if (NumRows(M) != NumCols(M))
      CoCoA_ERROR(ERR::NotSquareMatrix, "InverseByGauss");
    if (NumRows(lhs) != NumRows(M))
      CoCoA_ERROR(ERR::BadMatrixSize, "InverseByGauss");
    if (BaseRing(lhs) != BaseRing(M))
      CoCoA_ERROR(ERR::MixedRings, "InverseByGauss");
    matrix Gss(NewDenseMatrix(BaseRing(lhs), NumRows(lhs), NumCols(lhs)));
    // Copy M into Gss -- DenseMatrix Gss(NewDenseMatrix(M));
    for (size_t i=0; i < NumRows(lhs); ++i)
      for (size_t j=0; j < NumCols(lhs); ++j)
        SetEntry(Gss, i, j, M(i, j));

    // Set lhs to identity matrix --  lhs = IdentityMatrix(NumRows);
    lhs->myAssignZero();
    for (size_t i=0; i < NumRows(lhs); ++i)
      SetEntry(lhs,i,i,1);
    //    DenseMatrix Gss(M);

    RingElem c(BaseRing(lhs));
    for (size_t j=0; j < NumCols(lhs); ++j)
    {
      if (IsZero(Gss(j,j)))
      {
        size_t i=j+1;
        for ( ; i < NumRows(lhs); ++i)
          if (!IsZero(Gss(i,j))) break;
        if (i==NumRows(lhs)) CoCoA_ERROR(ERR::NotInvMatrix, "inverse");
        Gss->mySwapRows(i,j);
        lhs->mySwapRows(i,j);
      }
      c = 1/Gss(j,j);
      Gss->myRowMul(j, c);
      lhs->myRowMul(j, c);
      for (size_t i=0; i < NumRows(lhs); ++i)
        if (i!=j)
        {
          c = -Gss(i,j);
          Gss->myAddRowMul(i, j, c); // AddRowMul(Gss, i, j, c);
          lhs->myAddRowMul(i, j, c);
        }
    }
  }


  void AdjointByInverse(matrix& lhs, ConstMatrix M)
  {
    // This code works only if the matrix is invertible!
    // It is morally equivalent to swap(lhs, inverse(M)*det(M));
    CoCoA_ASSERT(IsIntegralDomain(BaseRing(M)));
    CoCoA_ASSERT(NumRows(M) == NumRows(M));
    if (IsField(BaseRing(M)))
    {
      matrix tmp(NewDenseMatrix(BaseRing(M), NumRows(M), NumCols(M)));
      inverse(tmp, M);
      RingElem d(det(tmp));
      for (size_t i=0; i < NumRows(M); ++i)
        for (size_t col=0; col < NumCols(M); ++col)
          SetEntry(tmp, i, col, tmp(i,col) * d);
      tmp->mySwap(lhs);
      return;
    }
    FractionField K(NewFractionField(BaseRing(M)));
    matrix tmp(NewDenseMatrix(K, NumRows(M), NumCols(M)));
    // below copy M via RingHom into K
    RingHom R2K(EmbeddingHom(K));
    for (size_t row=0; row < NumRows(M); ++row)
      for (size_t col=0; col < NumCols(M); ++col)
        SetEntry(tmp, row, col, R2K(M(row,col)));
  }



  bool IsTermOrdering(ConstMatrix M)
  {
    for (size_t col=0; col<NumCols(M); ++col)
    {
      size_t row=0;
      for (; row<NumRows(M); ++row)
      {
        if ( IsZero(M(row,col)) ) continue;
        if ( sign(M(row,col))<0 ) return false;
        else break;
      }
      if (row==NumRows(M)) return false;
    }
    return true;
  }


/**
    expects a matrix with entries in an ordered ring: very likely only Z (maybe Q?)
    returns a matrix with positive entries which defines an equivalent ordering
*/

  matrix NewPositiveMatrix(ConstMatrix M)
  {
    CoCoA_ASSERT( IsZ(BaseRing(M)) );
    if ( !IsTermOrdering(M) )
      CoCoA_ERROR(ERR::NotTermOrdering, "PositiveMatrix");

    matrix PM(NewDenseMatrix(M));
    RingElem one(BaseRing(PM),1);
    for (size_t row=0; row<NumRows(PM); ++row)
      for (size_t col=0; col<NumCols(PM); ++col)
        if (sign(PM(row,col))<0)
        {
          size_t PosRow=0;
          while (IsZero(PM(PosRow,col))) ++PosRow; // find positive entry in col
          while (sign(PM(row,col))<0) PM->myAddRowMul(row, PosRow, one);
        }
    return PM;
  }

/**
    expects a matrix with entries in an ordered ring
*/

  bool IsPositiveGrading(ConstMatrix M, size_t GradingDim)
  {
    if (!IsTermOrdering(M))  return false; // seems excessive but it isn't
    // check there are no null columns
    size_t i;
    for (size_t j=0; j < NumCols(M); ++j)
    {
      for (i=0; i < GradingDim; ++i) if (M(i,j) > 0) break;
      if (i==GradingDim) return false;
    }
    return true;
  }


  matrix NewMatrixConcatHor(ConstMatrix M1, ConstMatrix M2)
  {
    if ( BaseRing(M1)!=BaseRing(M2) )
      CoCoA_ERROR(ERR::MixedRings, "NewMatrixConcatHor");
    if ( NumRows(M1)!=NumRows(M2) )
      CoCoA_ERROR(ERR::BadMatrixSize, "NewMatrixConcatHor");

    matrix MM(NewDenseMatrix(BaseRing(M1), NumRows(M1), NumCols(M1)+NumCols(M2)));
    for (size_t row=0; row<NumRows(M1); ++row)
      for (size_t col=0; col<NumCols(M1); ++col)
        SetEntry(MM, row, col, M1(row, col));
    for (size_t row=0; row<NumRows(M2); ++row)
      for (size_t col=0, c=NumCols(M1)+1; col<NumCols(M2); ++col, ++c)
        SetEntry(MM, row, c, M2(row, col));
    return MM;
  }


  matrix NewMatrixConcatVer(ConstMatrix M1, ConstMatrix M2)
  {
    if ( BaseRing(M1)!=BaseRing(M2) )
      CoCoA_ERROR(ERR::MixedRings, "NewMatrixConcatVer");
    if ( NumCols(M1)!=NumCols(M2) )
      CoCoA_ERROR(ERR::BadMatrixSize, "NewMatrixConcatVer");

    matrix MM(NewDenseMatrix(BaseRing(M1), NumRows(M1)+NumRows(M2), NumCols(M1)));
    for (size_t row=0; row<NumRows(M1); ++row)
      for (size_t col=0; col<NumCols(M1); ++col)
        SetEntry(MM, row, col, M1(row, col));
    for (size_t row=0, r=NumRows(M1); row<NumRows(M2); ++row, ++r)
      for (size_t col=0; col<NumCols(M2); ++col)
        SetEntry(MM, r, col, M2(row, col));
    return MM;
  }


  matrix NewMatrixConcatDiag(ConstMatrix M1, ConstMatrix M2)
  {
    if ( BaseRing(M1)!=BaseRing(M2) )
      CoCoA_ERROR(ERR::MixedRings, "NewMatrixConcatDiag");

    matrix MM(NewDenseMatrix(BaseRing(M1), NumRows(M1)+NumRows(M2), NumCols(M1)+NumCols(M2)));
    for (size_t row=0; row<NumRows(M1); ++row)
      for (size_t col=0; col<NumCols(M1); ++col)
        SetEntry(MM, row, col, M1(row, col));
    for (size_t row=0, r=NumRows(M1)+1; row<NumRows(M2); ++row, ++r)
      for (size_t col=0, c=NumCols(M1)+1; col<NumCols(M2); ++col, ++c)
        SetEntry(MM, r, c, M2(row, col));
    return MM;
  }


  matrix NewMatrixConcatAntiDiag(ConstMatrix M1, ConstMatrix M2)
  {
    if ( BaseRing(M1)!=BaseRing(M2) )
      CoCoA_ERROR(ERR::MixedRings, "NewMatrixConcatAntiDiag");

    matrix MM(NewDenseMatrix(BaseRing(M1), NumRows(M1)+NumRows(M2), NumCols(M1)+NumCols(M2)));
    for (size_t row=0, r=NumRows(M2)+1; row<NumRows(M1); ++row, ++r)
      for (size_t col=0; col<NumCols(M1); ++col)
        SetEntry(MM, r, col, M1(row, col));
    for (size_t row=0; row<NumRows(M2); ++row)
      for (size_t col=0, c=NumCols(M1)+1; col<NumCols(M2); ++col, ++c)
        SetEntry(MM, row, c, M2(row, col));
    return MM;
  }


  matrix NewMatrixMinimize(ConstMatrix M)
  {
    if ( NumRows(M)<NumCols(M) )
      CoCoA_ERROR(ERR::BadMatrixSize, "MatrixMinimize");
    if ( rank(M)!=NumCols(M) )
      CoCoA_ERROR(ERR::NotFullRank, "MatrixMinimize");

    matrix MM(NewDenseMatrix(BaseRing(M), NumCols(M), NumCols(M)));
    size_t row=0;
    for (size_t r=0; r<NumRows(M); ++r)
    {
      for (size_t col=0; col<NumCols(M); ++col)
        SetEntry(MM, row, col, M(r, col));
      if (rank(MM) > row) ++row;
      if (row==NumRows(MM)) return MM;
    }
    CoCoA_ERROR(ERR::SERIOUS, "MatrixMinimize: should never get here");
    return MM;
  }


  matrix NewDenseMatrixRevLex(ring R, size_t n)
  {
    matrix RL(NewDenseMatrix(R,n,n));
    for (size_t r=0 ; r<n ; ++r)
      SetEntry(RL, r, n-r-1, -1);
    return RL;
  }


  matrix NewMatrixCompleteOrd(ConstMatrix M)
  {
    matrix MC(NewMatrixConcatVer(M, NewDenseMatrixRevLex(BaseRing(M), NumCols(M))));
    return NewMatrixMinimize(MC);
  }

  matrix NewMatrixElim(ConstMatrix GradM,
                       std::vector<size_t> IndetsToElim,
		       const bool IsHomog)
  {
    if (NumRows(GradM)==0)
      return NewMatrix0Elim(GradM,IndetsToElim);
    if (IsHomog)
      return NewMatrixHElim(GradM,IndetsToElim);
    return  NewMatrixAElim(GradM,IndetsToElim);
  }

  matrix NewMatrixElim(size_t NumIndets, std::vector<size_t> IndetsToElim)
  {
    matrix ElimRows(NewDenseMatrix(RingZ(), 2, NumIndets));

    for (size_t i=0 ; i<IndetsToElim.size() ; ++i)
      SetEntry(ElimRows, 0, IndetsToElim[i], 1);
    for (size_t i=0 ; i<NumIndets ; ++i)
      if ( ElimRows(0,i)==0 )  SetEntry(ElimRows, 1, i, 1);
    return NewMatrixCompleteOrd(ElimRows);
  }


  // this should be fixed: it may give a non-term-ordering
  matrix NewMatrix0Elim(ConstMatrix GradM,std::vector<size_t> IndetsToElim)
  {
    matrix ElimRow(NewDenseMatrix(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0 ; i<IndetsToElim.size() ; ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);

    matrix OneRow(NewDenseMatrix(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0 ; i<NumCols(GradM) ; ++i)
      SetEntry(OneRow, 0, i, 1);
    return NewMatrixCompleteOrd(NewMatrixConcatVer(ElimRow,OneRow));
  }

  matrix NewMatrixHElim(ConstMatrix GradM, std::vector<size_t> IndetsToElim)
  {
    matrix ElimRow(NewDenseMatrix(BaseRing(GradM), 1, NumCols(GradM)));
    for (size_t i=0 ; i<IndetsToElim.size() ; ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);
    return NewMatrixCompleteOrd(NewMatrixConcatVer(GradM, ElimRow));
  }

  matrix NewMatrixAElim(ConstMatrix GradM, std::vector<size_t> IndetsToElim)
  {
    matrix ElimRow(NewDenseMatrix(BaseRing(GradM), 1, NumCols(GradM)));

    for (size_t i=0 ; i<IndetsToElim.size() ; ++i)
      SetEntry(ElimRow, 0, IndetsToElim[i], 1);
    return NewMatrixCompleteOrd(NewMatrixConcatVer(ElimRow,GradM));
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/MatrixArith.C,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: MatrixArith.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.10  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.9  2007/03/05 21:06:07  cocoa
// New names for homomorphism pseudo-ctors: removed the "New" prefix.
//
// Revision 1.8  2007/03/02 10:47:53  cocoa
// First stage of RingZ modifications -- tests do not compile currently, Anna will fix this.
//
// Revision 1.7  2007/02/10 18:44:03  cocoa
// Added "const" twice to each test and example.
// Eliminated dependency on io.H in several files.
// Improved BuildInfo, and added an example about how to use it.
// Some other minor cleaning.
//
// Revision 1.6  2006/12/21 13:48:32  cocoa
// Made all increment/decrement calls prefix (except where the must be postfix).
//
// Revision 1.5  2006/11/27 16:18:32  cocoa
// -- moved classes declarations from .H to .C (DenseMatrix, DiagMatrix,
//    FieldIdeal, SpecialMatrix)
//
// Revision 1.4  2006/10/06 14:04:15  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.3  2006/08/17 09:39:07  cocoa
// -- added: elimination ordering matrix for non-homogeneous input
//
// Revision 1.2  2006/07/17 16:58:05  cocoa
// -- added: NewMatrixElim(size_t NumIndets, std::vector<size_t> IndetsToElim)
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.8  2006/05/02 14:39:20  cocoa
// -- Changed "not" into "!" becuase of M$Windoze (by M.Abshoff)
//
// Revision 1.7  2006/04/27 13:45:30  cocoa
// Changed name of NewIdentityRingHom to NewIdentityHom.
// Changed name of member functions which print out their own object
// into myOutputSelf (to distinguish from "transitive" myOutput fns).
//
// Revision 1.6  2006/04/10 13:20:43  cocoa
// -- fixed buglets for Elimination orderings
//
// Revision 1.5  2006/04/05 16:45:29  cocoa
// -- added comment and CoCoA_ASSERT in NewPositiveMatrix
// -- added IsPositiveGrading
//
// Revision 1.4  2006/04/05 14:49:20  cocoa
// -- fixed: NewPositiveMatrix (tested and used in OrdvArith.C)
//
// Revision 1.3  2006/01/19 15:48:49  cocoa
// -- fixed RankByGauss by Stefan Kaspar
//
// Revision 1.2  2005/12/31 12:22:17  cocoa
// Several minor tweaks to silence the Microsoft compiler:
//  - added some missing #includes and using directives
//  - moved some function defns into the right namespace
//  - etc.
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.5  2005/04/20 15:40:48  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.4  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.3  2005/04/19 14:06:04  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.2  2005/03/30 17:15:14  cocoa
// Cleaned the SpecialMatrix code; a simple test compiles and
// runs fine.
//
// Revision 1.1  2005/03/11 18:38:32  cocoa
// -- first import
//
