//   Copyright (c)  2005  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License (version 2)
//   as published by the Free Software Foundation.  A copy of the full
//   licence may be found in the file COPYING in this directory.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoA; if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#include "CoCoA/SpecialMatrix.H"
#include "CoCoA/assert.H"
#include "CoCoA/matrix.H"
#include "CoCoA/ring.H"
#include "CoCoA/MatrixArith.H"  // for RankByGauss


//#include <cstddef>
using std::size_t;
//#include <vector>
using std::vector;
#include <algorithm>
using std::min;

namespace CoCoA
{

  /***************************************************************************/
  // ZeroMatrixImpl -- zero matrices

  class ZeroMatrixImpl: public ConstMatrixBase
  {
  private:
    friend ConstMatrix NewZeroMatrix(const ring&, std::size_t, std::size_t); // pseudo-ctor
    ZeroMatrixImpl(const ring& R, std::size_t nrows, std::size_t ncols);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    ZeroMatrixImpl(const ZeroMatrixImpl&);            // NEVER DEFINED -- copy ctor disabled
    ZeroMatrixImpl& operator=(const ZeroMatrixImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
  };


  /***************************************************************************/
  // IdentityMatrixImpl -- identity matrices (necessarily square)

  class IdentityMatrixImpl: public ConstMatrixBase
  {
  private:
    friend ConstMatrix NewIdentityMatrix(const ring&, std::size_t); // pseudo-ctor
    IdentityMatrixImpl(const ring& R, std::size_t dim);
    // default dtor is fine
  private: // disable default copy ctor and assignment
    IdentityMatrixImpl(const IdentityMatrixImpl&);            // NEVER DEFINED -- copy ctor disabled
    IdentityMatrixImpl& operator=(const IdentityMatrixImpl&); // NEVER DEFINED -- copy ctor disabled
  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    size_t myDim;
  };


  /***************************************************************************/
  // TransposeMatrixImpl -- a transposed view of a matrix

  class TransposeMatrixImpl: public ConstMatrixBase
  {
  private:
    friend ConstMatrix NewTransposeMatrix(ConstMatrix M); // pseudo-ctor
    TransposeMatrixImpl(ConstMatrix M);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    TransposeMatrixImpl(const TransposeMatrixImpl&);            // NEVER DEFINED -- copy ctor disabled
    TransposeMatrixImpl& operator=(const TransposeMatrixImpl&); // NEVER DEFINED -- copy ctor disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
    virtual void myMulByRow(vec& lhs, const vec& v) const;
    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
    virtual void myDet(RefRingElem d) const;
    virtual std::size_t myRank() const;

  private: // data members
    ConstMatrix myM;
  };


  /***************************************************************************/
  // SubmatrixImpl -- see only certain rows/cols of a matrix

  class SubmatrixImpl: public ConstMatrixBase
  {
  private:
    friend ConstMatrix NewSubmatrix(ConstMatrix, const std::vector<std::size_t>&, const std::vector<std::size_t>&); // pseudo-ctor
    SubmatrixImpl(ConstMatrix M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    SubmatrixImpl(const SubmatrixImpl&);            // NEVER DEFINED -- copy ctor disabled
    SubmatrixImpl& operator=(const SubmatrixImpl&); // NEVER DEFINED -- assignment disabled

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use DEFAULT definitions of the following six functions.
//     virtual void myMulByRow(vec& lhs, const vec& v) const;
//     virtual void myMulByCol(vec& lhs, const vec& v) const;
//     virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
//     virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
//     virtual void myDet(RefRingElem d) const;
//     virtual std::size_t myRank() const;

  private: // data members
    ConstMatrix myM;
    std::vector<std::size_t> myRowTable;
    std::vector<std::size_t> myColTable;
  };


  class BlockMatrixImpl: public ConstMatrixBase
  {
  private:
    friend ConstMatrix NewBlockMatrix(ConstMatrix A, ConstMatrix B, ConstMatrix C, ConstMatrix D); // pseudo-ctor
    BlockMatrixImpl(ConstMatrix A, ConstMatrix B, ConstMatrix C, ConstMatrix D);
    // default dtor works fine
  private: // disable default copy ctor and assignment
    BlockMatrixImpl(const BlockMatrixImpl&);            // NEVER DEFINED -- disable copy ctor
    BlockMatrixImpl& operator=(const BlockMatrixImpl&); // NEVER DEFINED -- disable assignment

  public:
    typedef std::vector<RingElem> vec;
    const ring& myBaseRing() const;
    virtual std::size_t myNumRows() const;
    virtual std::size_t myNumCols() const;
    virtual ConstRefRingElem myEntry(std::size_t i, std::size_t j) const;
// Use default defns of the following two functions
//    virtual void myMulByRow(vec& lhs, const vec& v) const;
//    virtual void myMulByCol(vec& lhs, const vec& v) const;
    virtual bool myIsZeroRow(std::size_t i) const; ///< tests whether row i is zero
    virtual bool myIsZeroCol(std::size_t j) const; ///< tests whether col j is zero
// Use default defns of the following two functions
//    virtual void myDet(RefRingElem d) const;
//    virtual std::size_t myRank() const;

  private: // data members
    const ring myR;
    std::size_t myNumRowsValue;
    std::size_t myNumColsValue;
    ConstMatrix myA, myB, myC, myD;
  };


  //----------------------------------------------------------------------
  // Functions for ZeroMatrixImpl

  ZeroMatrixImpl::ZeroMatrixImpl(const ring& R, std::size_t NumRows, std::size_t NumCols):
      myR(R),
      myNumRowsValue(NumRows),
      myNumColsValue(NumCols)
  {}


  const ring& ZeroMatrixImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t ZeroMatrixImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t ZeroMatrixImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem ZeroMatrixImpl::myEntry(std::size_t /*i*/, std::size_t /*j*/) const
  {
    return zero(myR);
  }


  // This is exception safe only if "lhs[i]=0;" cannot throw
  void ZeroMatrixImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    for (size_t i=0; i < myNumCols(); ++i)
      lhs[i] = 0;
  }


  // This is exception safe only if "lhs[i]=0;" cannot throw
  void ZeroMatrixImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    for (size_t i=0; i < myNumRows(); ++i)
      lhs[i] = 0;
  }


  bool ZeroMatrixImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return true;
  }


  bool ZeroMatrixImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return true;
  }


  void ZeroMatrixImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumRows());
    d = 0;
  }



  size_t ZeroMatrixImpl::myRank() const
  {
    return 0;
  }



  ConstMatrix NewZeroMatrix(const ring& R, std::size_t NumRows, std::size_t NumCols)
  {
    return ConstMatrix(new ZeroMatrixImpl(R, NumRows, NumCols));
  }




  /***************************************************************************/
  // Functions for IdentityMatrixImpl

  IdentityMatrixImpl::IdentityMatrixImpl(const ring& R, size_t dim):
      myR(R),
      myDim(dim)
  {}


  const ring& IdentityMatrixImpl::myBaseRing() const
  {
    return myR;
  }


  size_t IdentityMatrixImpl::myNumRows() const
  {
    return myDim;
  }


  size_t IdentityMatrixImpl::myNumCols() const
  {
    return myDim;
  }


  ConstRefRingElem IdentityMatrixImpl::myEntry(size_t i, size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows());
    CoCoA_ASSERT(j < myNumCols());
    if (i == j) return one(myR);
    return zero(myR);
  }


  void IdentityMatrixImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
//???    CoCoA_ASSERT(myBaseRing() == AmbientRing(v));
    lhs = v;
  }


  void IdentityMatrixImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
//???    CoCoA_ASSERT(myBaseRing() == AmbientRing(v));
    lhs = v;
  }


  bool IdentityMatrixImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myDim);
    return false;
  }


  bool IdentityMatrixImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myDim);
    return false;
  }


  void IdentityMatrixImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    if (myNumRows() == 0)
      d = 0;
    else
      d = 1;
  }


  size_t IdentityMatrixImpl::myRank() const
  {
    return myNumRows();
  }


  ConstMatrix NewIdentityMatrix(const ring& R, size_t dim)
  {
    return ConstMatrix(new IdentityMatrixImpl(R, dim));
  }



  /***************************************************************************/
  // Functions for TransposeMatrixImpl -- all are simple one-liners

  TransposeMatrixImpl::TransposeMatrixImpl(ConstMatrix M):
      myM(M)
  {}


  const ring& TransposeMatrixImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t TransposeMatrixImpl::myNumRows() const
  {
    return NumCols(myM);
  }


  std::size_t TransposeMatrixImpl::myNumCols() const
  {
    return NumRows(myM);
  }


  ConstRefRingElem TransposeMatrixImpl::myEntry(std::size_t i, std::size_t j) const
  {
    return myM(j, i);
  }


  void TransposeMatrixImpl::myMulByRow(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumCols() && v.size() == myNumRows());
    myM->myMulByCol(lhs, v);
  }


  void TransposeMatrixImpl::myMulByCol(vec& lhs, const vec& v) const
  {
    CoCoA_ASSERT(lhs.size() == myNumRows() && v.size() == myNumCols());
    myM->myMulByRow(lhs, v);
  }



  bool TransposeMatrixImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    return myM->myIsZeroCol(i);
  }


  bool TransposeMatrixImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    return myM->myIsZeroRow(j);
  }


  void TransposeMatrixImpl::myDet(RefRingElem d) const
  {
    CoCoA_ASSERT(owner(d) == myBaseRing());
    CoCoA_ASSERT(myNumRows() == myNumRows());
    myM->myDet(d);
  }


  size_t TransposeMatrixImpl::myRank() const
  {
    return myM->myRank();
  }


  ConstMatrix NewTransposeMatrix(ConstMatrix M)
  {
    return ConstMatrix(new TransposeMatrixImpl(M));
  }


  /***************************************************************************/
  // Functions for SubmatrixImpl


  SubmatrixImpl::SubmatrixImpl(ConstMatrix M, const std::vector<std::size_t>& rows, const std::vector<std::size_t>& cols):
      myM(M),
      myRowTable(rows),
      myColTable(cols)
  {
    // Args already checked by pseudo-ctor NewSubmatrix, the only
    // function which calls this ctor
  }


  const ring& SubmatrixImpl::myBaseRing() const
  {
    return BaseRing(myM);
  }


  std::size_t SubmatrixImpl::myNumRows() const
  {
    return myRowTable.size();
  }


  std::size_t SubmatrixImpl::myNumCols() const
  {
    return myColTable.size();
  }


  ConstRefRingElem SubmatrixImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    return myM(myRowTable[i], myColTable[j]);
  }


  ConstMatrix NewSubmatrix(ConstMatrix M, const std::vector<std::size_t>& rows, const vector<std::size_t>& cols)
  {
    // Check all row indices are within range.
    const size_t r = NumRows(M);
    for (size_t i=0; i < rows.size(); ++i)
      if (rows[i] >= r)
        CoCoA_ERROR(ERR::BadRowIndex, "NewSubmatrix");

    // Check all col indices are within range.
    const size_t c = NumCols(M);
    for (size_t j=0; j < cols.size(); ++j)
      if (cols[j] >= c)
        CoCoA_ERROR(ERR::BadColIndex, "NewSubmatrix");

    // Check all row indices are distinct.
    for (size_t i=0; i < rows.size(); ++i)
      for (size_t j=i+1; j < rows.size(); ++j)
        if (rows[i] == rows[j])
          CoCoA_ERROR(ERR::BadRowIndex, "NewSubmatrix (repeated row indices)");

    // Check all col indices are distinct.
    for (size_t i=0; i < cols.size(); ++i)
      for (size_t j=i+1; j < cols.size(); ++j)
        if (cols[i] == cols[j])
          CoCoA_ERROR(ERR::BadColIndex, "NewSubmatrix (repeated col indices)");

    return ConstMatrix(new SubmatrixImpl(M, rows, cols));
  }


  /***************************************************************************/
  // Functions for BlockMatrixImpl -- a matrix of the form  (A B)
  //                                                        (C D)
  // This implementation requires that the submatrices correspond
  // to two straight cuts -- several changes would be needed to handle
  // a more general assembly of submatrices.

  BlockMatrixImpl::BlockMatrixImpl(ConstMatrix A, ConstMatrix B, ConstMatrix C, ConstMatrix D):
      myR(BaseRing(A)),
      myNumRowsValue(NumRows(A)+NumRows(C)),
      myNumColsValue(NumCols(A)+NumCols(B)),
      myA(A),
      myB(B),
      myC(C),
      myD(D)
  {
    // Args are checked for compatibility by NewBlockMatrix, the
    // only function which calls this ctor.
  }


  const ring& BlockMatrixImpl::myBaseRing() const
  {
    return myR;
  }


  std::size_t BlockMatrixImpl::myNumRows() const
  {
    return myNumRowsValue;
  }


  std::size_t BlockMatrixImpl::myNumCols() const
  {
    return myNumColsValue;
  }


  ConstRefRingElem BlockMatrixImpl::myEntry(std::size_t i, std::size_t j) const
  {
    CoCoA_ASSERT(i < myNumRows() && j < myNumCols());
    if (i < NumRows(myA) && j < NumCols(myA))
      return myA(i, j);
    if (i < NumRows(myB))
      return myB(i, j-NumCols(myA));
    if (j < NumCols(myC))
      return myC(i-NumRows(myA), j);
    return myD(i-NumRows(myA), j-NumCols(myA));
  }


  bool BlockMatrixImpl::myIsZeroRow(std::size_t i) const
  {
    CoCoA_ASSERT(i < myNumRows());
    if (i < NumRows(myA))
      return myA->myIsZeroRow(i) && myB->myIsZeroRow(i);
    return myC->myIsZeroRow(i-NumRows(myA)) && myD->myIsZeroRow(i-NumRows(myA));
  }


  bool BlockMatrixImpl::myIsZeroCol(std::size_t j) const
  {
    CoCoA_ASSERT(j < myNumCols());
    if (j < NumCols(myA))
      return myA->myIsZeroCol(j) && myC->myIsZeroCol(j);
    return myB->myIsZeroCol(j-NumCols(myA)) && myD->myIsZeroCol(j-NumCols(myA));
  }



  ConstMatrix NewBlockMatrix(ConstMatrix A, ConstMatrix B, ConstMatrix C, ConstMatrix D)
  {
    {
      // Check that the rings are all the same
      const ring R = BaseRing(A);
      if (BaseRing(B) != R || BaseRing(C) != R || BaseRing(D) != R)
        CoCoA_ERROR(ERR::MixedRings, "NewBlockMatrix(A,B,C,D)");
    }
    // Check that the matrix sizes are suitably matched.
    if (NumRows(A) != NumRows(B) ||
        NumRows(C) != NumRows(D) ||
        NumCols(A) != NumCols(C) ||
        NumCols(B) != NumCols(D))
      CoCoA_ERROR(ERR::BadMatrixSize, "NewBlockMatrix(A,B,C,D)");
    return ConstMatrix(new BlockMatrixImpl(A, B, C, D));
  }



} // End of namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/SpecialMatrix.C,v 1.1.1.1 2007/03/09 15:16:11 abbott Exp $
// $Log: SpecialMatrix.C,v $
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.5  2007/03/08 18:22:29  cocoa
// Just whitespace cleaning.
//
// Revision 1.4  2007/03/07 22:25:18  bigatti
// -- commented out unused parameters (for -Wextra)
//
// Revision 1.3  2006/11/27 16:18:32  cocoa
// -- moved classes declarations from .H to .C (DenseMatrix, DiagMatrix,
//    FieldIdeal, SpecialMatrix)
//
// Revision 1.2  2006/10/06 14:04:14  cocoa
// Corrected position of #ifndef in header files.
// Separated CoCoA_ASSERT into assert.H from config.H;
// many minor consequential changes (have to #include assert.H).
// A little tidying of #include directives (esp. in Max's code).
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.8  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.7  2005/04/19 15:39:55  cocoa
// Matrices now use reference counts.
//
// Revision 1.6  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.5  2005/03/31 16:59:42  cocoa
// Made special matrix ctors private, so a user has to pass via the
// pseudo-ctors (which do all the arg sanity checking).
//
// Revision 1.4  2005/03/30 17:15:14  cocoa
// Cleaned the SpecialMatrix code; a simple test compiles and
// runs fine.
//
